/*#io
SocketManager ioDoc(
docCopyright("Steve Dekorte", 2004)
docLicense("BSD revised")
docObject("SocketManager")
docDescription("The SocketManager manages all socket objects waiting on read, writes or connects. When a socket does one of these operations, it adds it's descriptor to the SockManager's descriptor list and the SocketManager calls select() on this list in it's own coroutine and informs the sockets when the event they are waiting on occurs. The socket manager's coroutine automatically starts when it gets a descriptor and stops when it has no descriptors to process.")
*/

#include "IoSocketManager.h"
#include "IoState.h"
#include "IoObject_actor.h"
#include "IoNil.h"
#include "IoNumber.h"
#include "Scheduler.h"

#define DEBUG_IOSOCKETMANANGER 1

#define DATA(self) ((IoSocketManagerData *)self->data)

IoSocketManager *IoSocketManager_socketManagerForState_(void *state)
{
    IoSocketManager *sm =  IoState_protoWithName_(state, "SocketManager");
    /*
     IoSocketManager *sm = IoObject_getSlot_(((IoState *)state)->lobby, 
					     IoState_symbolWithCString_(state, "SocketManager"));
     */
    if (ISSOCKETMANAGER(sm)) 
    {
	return sm;
    }
    
    IoState_error_description_(state, IoMessage_newWithName_(state,  
	IoState_symbolWithCString_(state, "Socket")), 
	"Socket.clone", 
	"Unable to find SocketManager proto");
    return sm;
}

IoTag *IoSocketManager_tag(void *state)
{
    IoTag *tag = IoTag_newWithName_("SocketManager");
    tag->state = state;
    tag->cloneFunc = (TagCloneFunc *)IoSocketManager_rawClone; 
    tag->freeFunc  = (TagFreeFunc *)IoSocketManager_free;
    tag->markFunc  = (TagMarkFunc *)IoSocketManager_mark;
    return tag;
}

IoSocketManager *IoSocketManager_proto(void *state)
{
    IoObject *self = IoObject_new(state);
    self->tag = IoSocketManager_tag(state);
    
    self->data = calloc(1, sizeof(IoSocketManagerData));    
    DATA(self)->socketManager = SocketManager_new();
    
    SocketManager_scheduler_(DATA(self)->socketManager, IOSTATE->scheduler);
    SocketManager_target_(DATA(self)->socketManager, self);
    SocketManager_callback_(DATA(self)->socketManager, (SocketManagerCallback *)IoSocketManager_update);
    
    /*SocketManager_scheduler_(DATA(self)->socketManager, ((IoState *)state)->scheduler);*/
    
    IoState_registerProtoWithFunc_(state, self, IoSocketManager_proto);
    
    {
	IoMethodTable methodTable[] = 
	{
	    {"descriptorCount", IoSocketManager_descriptorCount},
	    {"setSelectTimeout", IoSocketManager_setSelectTimeout},
	    {"selectTimeout", IoSocketManager_selectTimeout},
	    {"run", IoSocketManager_run},
	    {NULL, NULL},
	};
	IoObject_addMethodTable_(self, methodTable);
    }
    return self;
}

/* ----------------------------------------------------------- */

IoSocketManager *IoSocketManager_rawClone(IoSocketManager *self)
{ 
    return self; 
}

void IoSocketManager_mark(IoSocketManager *self) 
{ 
    /*
    SocketManager *sm = DATA(self)->socketManager;
    List_do_(sm->readDescriptors, (ListDoCallback *)IoSocketManager_markDescriptor);
    List_do_(sm->writeDescriptors, (ListDoCallback *)IoSocketManager_markDescriptor);
    */
}

void IoSocketManager_free(IoSocketManager *self) 
{ 
    SocketManager_free(DATA(self)->socketManager);
    free(self->data); 
}

SocketManager *IoSocketManager_rawSocketManager(IoSocketManager *self)
{ 
    return DATA(self)->socketManager; 
}

void IoSocketManager_update(IoSocketManager *self)
{
    /* create/resume coro when there are descriptors to process, 
    * pause it when there aren't any left
    */
    if (SocketManager_descriptorCount(DATA(self)->socketManager))
    {
	if (!self->coro)
	{
	    IoMessage *asyncMessage = IoMessage_newWithName_label_(IOSTATE, 
		IOSYMBOL("@"), IOSYMBOL("IoSocketManager"));
	    IoMessage *runMessage = IoMessage_newWithName_label_(IOSTATE, 
		IOSYMBOL("run"), IOSYMBOL("IoSocketManager"));
	    IoMessage_addArg_(asyncMessage, runMessage);
	    IoObject_asyncPerform(self, self, asyncMessage);
	    Coro_label_(self->coro->cid, "IoSocketManager");
	    IoState_yield(IOSTATE);
	}
	else
	{ 
	    IoObject_resume(self);
	}
    }
    else
    {
	IoObject_prepareToPause(self);
    }
}

/* ----------------------------------------------------------- */

IoObject *IoSocketManager_descriptorCount(IoSocketManager *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("descriptorCount", 
    "Returns number of descriptors currently being watched by the receiver. ")
    */
    
    return IONUMBER(SocketManager_descriptorCount(DATA(self)->socketManager)); 
}

IoObject *IoSocketManager_setSelectTimeout(IoSocketManager *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("setSelectTimeout(seconds)", 
    "Set the number of seconds to wait in select() for new events. This number can (and generally should) be a fraction. Example:
<pre>
  SocketManager setSelectTimeout(.05) // wait for 1/20 of a second
</pre>")
    */
    
    IoNumber *seconds = IoMessage_locals_numberArgAt_(m, locals, 0);
    SocketManager_timeoutSeconds_(DATA(self)->socketManager, IoNumber_asFloat(seconds));
    return self; 
}

IoObject *IoSocketManager_selectTimeout(IoSocketManager *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("selectTimeout", 
    "Returns the select timeout.")
    */
    
    float v = SocketManager_timeoutSeconds(DATA(self)->socketManager);
    return IONUMBER(v); 
}

IoObject *IoSocketManager_run(IoSocketManager *self, IoObject *locals, IoMessage *m)
{
    IoState *state = IOSTATE;
    Scheduler_yield(state->scheduler);
    
    for(;;) 
    {
	/* 
	the coro will be paused/resumed when appropiate 
	in the IoSocketManager_update() callback 
	*/
	SocketManager_timer(DATA(self)->socketManager);
	IoState_yield(state);
    }
    
    return self;
}
