/*   
copyright: Steve Dekorte, 2004. All rights reserved.
license: See _BSDLicense.txt.

object: URL
prototype: Object

module: IoServer
description: The URL object is usefull for fetching web pages and parsing URLs. Example;
<pre>
page := URL clone setURL("http://www.google.com/") fetch
</pre>
*/


Sequence parseAsUrlToMap := method(
    m := Map clone
    s := self
    
    protocol := s beforeSeq("://")
    s := s afterSeq("://")
    if (s == Nil, return)
    host := s beforeSeq("/")
    if (host containsSeq(":")) then (
      port := host afterSeq(":")
      host := host  beforeSeq(":")
    ) else (
      port := "80"
    )

    request := s afterSeq("/")

    m atPut("protocol", protocol)
    m atPut("host", host)
    m atPut("port", port)
    m atPut("request", request)
    
    m
)

URL := Object clone do(
     type := "URL"

    /*
    method: url
    description: Returns url string.
    */
    url := ""

    /*
    method: setURL(urlString)
    description: Sets the url string and parses into the protocol, host, port path, and query slots. Returns self.
    */
    setURL := method(s, 
	url = s asString
	parse
	self
    )
    
    /*
    method: error
    description: Returns error string if it has been set or Nil otherwise.
    */
    newSlot("error", Nil)
    
    clear := method(
	self do(
	protocol := Nil
	host  := Nil
	port  := 80
	path  := "/"
	query := Nil
	error := Nil
	)
    )

    /*
    method: escapeString(aString)
    description: Returns a new String that is aString with the appropriate characters replaced by their URL escape codes.
    */

    escapeString := method(u,
	u := u clone asSequence 
	u replaceSeq("%","%25")
	escapeCodes foreach(k, v, u replaceSeq(k, v))
	u asString
    )

    /*
    method: unescapeString(aString)
    description: Returns a new String that is aString with the URL escape codes replaced by the appropriate characters.
    */
    
    unescapeString := method(u,
	u := u clone asSequence 
	escapeCodes foreach(k, v, u replaceSeq(v, k))
	u replaceSeq("%25", "%")
	u asString
    )

    /*
    method: escapeCodes
    description: Returns a Map whose key/value pairs are special characters and their URL escape codes.
    */
    
    escapeCodes := Map clone do(
        atPut(" ", "%20")
        atPut("<","%3C")
        atPut(">","%3E")
        atPut("#","%23")
        //atPut("%","%25")
        atPut("{","%7B")
        atPut("}","%7D")
        atPut("|","%7C")
        //atPut("backslash","%5C")
        atPut("^","%5E")
        atPut("~","%7E")
        atPut("[","%5B")
        atPut("]","%5D")
        atPut("\`","%60")
        atPut(";","%3B")
        atPut("/","%2F")
        atPut("?","%3F")
        atPut(":","%3A")
        atPut("@","%40")
        atPut("=","%3D")
        atPut("&","%26")
        atPut("$","%24")
    )

    /*
    method: referer 
    description: Returns the referer String or Nil if not set.
    */
    
    /*
    method: setReferer(aString)
    description: Sets the referer. Returns self.
    */
    newSlot("referer", Nil)

    parse := method(
	clear
	m := url parseAsUrlToMap 
	
	if (m at("request")) then (
	  m atPut("request", "/" .. m at("request"))
	) else (
	  m atPut("request", "/")
	)
	m atPut("path", m at("request"))
	m foreach(k, v, self setSlot(k, v); /*write(k, "=", v, "\n")*/)
	if (port) then (port = port asNumber) else (port = 80)
    )

    /*
    method: fetch
    description: Fetches the url and returns the result as a Sequence. Raises an exception if an error occurs.
    */
    
    fetch := method(url,
	if (url, setURL(url))
	if (protocol == "http", return fetchHttp)
	Exception raise("URL fetch", "protocol '" .. protocol .. "' unsupported")
    )

    /*
    method: fetchWithProgress(progressBlock)
    description: Same as fetch, but with each read, progressBlock is called with the  
    readBuffer and the content size as parameters.
    */

    fetchWithProgress := method(progressBlock,
	if (protocol == "http", return fetchHttpWithProgress(getSlot("progressBlock")))
	Exception raise("URL.fetch", "protocol '" .. protocol .. "' unsupported")
    )

    /*
    method: stopFetch
    description: Stops the fetch, if there is one. Returns self.
    */
    stopFetch := method(socket close; self)

    /*
    method: requestHeader
    description: Returns a Sequence containing the request header that will be sent.
    */
    requestHeader := method(
	header := Sequence clone
	//write("request = [", request, "]\n")
	header appendSeq("GET ", request," HTTP/1.0\r\n")
	header appendSeq("User-Agent: Mozilla/5.0\r\n")
	header appendSeq("Host: ", host, "\r\n")
	if (referer, header appendSeq("Referer: ", referer, "\r\n"))
	header appendSeq("Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, image/png, image/tiff, multipart/x-mixed-replace, */*\r\n")
	header appendSeq("Accept-Charset: iso-8859-1, utf-8, iso-10646-ucs-2, macintosh, windows-1252, *\r\n")
	header appendSeq("Accept-Language: en, *;q=0.5\r\n")
	header appendSeq("\r\n")
	header
    )

    fetchHttp := method(
	ip := Host clone setName(host) address
	socket := Socket clone setHost(ip) setPort(port) 
	socket connect 
	if (socket error, error := socket error; return Nil) 

	socket write(requestHeader)
	while(socket read, Nop)
	if (socket error) then (write(socket error, "\n"); return ) 
	/*write("read ", socket readBuffer size, " bytes\n")*/
	socket close

	b := socket readBuffer
	b clipBeforeEndOfSeq("\r\n\r\n") ifNil(b clipBeforeEndOfSeq("\n\n"))
	b
    )

    fetchHttpWithProgress := method(progressBlock,
	ip := Host clone setName(host) address
	socket := Socket clone setHost(ip) setPort(port)
	socket connect
	if (socket error, error := socket error; return Nil)

	socket write(requestHeader)

	socket read
	b := socket readBuffer
	self contentLength := 0
	beg := b findSeq( "Content-Length:" )
	if(beg,
	    end := b findSeq( "\r", beg + 16 )
	    self contentLength = b slice( beg + 16, end ) asNumber
	)

	while( socket read, progressBlock(b size, contentLength))
	if (socket error) then (write(socket error, "\n"); return)
	/*write("read ", socket readBuffer size, " bytes\n")*/
	socket close

	/*write("buffer[", b, "]\n")*/
	b clipBeforeEndOfSeq("\r\n\r\n") ifNil(b clipBeforeEndOfSeq("\n\n"))

	return b
    )

    /*
	method: fetchToFile(aFile)
	description: fetch the url and save the result to the specified File object. Saving is done as the data is read, which help minimize memory usage. Returns self on success or Nil on error.
    */
    fetchToFile := method(file,
	ip := Host clone setName(host) address
	socket := Socket clone setHost(ip) setPort(port) 
	socket connect 

	if (socket error, error := socket error; return Nil) 

	socket write(requestHeader)
	b := socket readBuffer
	
	while(socket read, 
	    if (b containsSeq("\r\n\r\n") or b containsSeq("\n\n"), break)
	)
	
	if (socket error) then ( write(socket error, "\n"); return )

	if (b clipBeforeEndOfSeq("\r\n\r\n") == Nil, b clipBeforeEndOfSeq("\n\n"))
	
	file write(b)
	b empty

	while(socket read, 
	    file write(b)
	    b empty
	)
	socket close
	
	if (socket error, self setError(socket error); return Nil)     
	self
    )


    childUrl := method(u, 
	if (u beginsWithSeq("http") == Nil, 
	    u = Path with(url pathComponent, u)
	)
	URL clone setURL(u) setReferer(url)
    )

    /*
    method: post(data)
    description: Sends an http post message. If data is a Map, it's key/value pairs are send as the post parameters. If data is a Sequence or String, it is sent directly. Returns self on success or Nil on error.
    */
    post := method(postdata,
	ip := Host clone setName(host) address

	header := Sequence clone
	header appendSeq(
		"POST ", request, " HTTP/1.0\r\n",
		"User-Agent: Mozilla/4.0\r\n",
		"Host: ", host, "\r\n",
		"Accept: */*\r\n",
		"Content-Type: application/x-www-form-urlencoded\r\n"
	)
	
	if (postdata type == "Map") then (
	    buffer := Sequence clone
	    postdata keys foreach(i, j,
		buffer appendSeq(escapeString(j), "=", escapeString(postdata at(j)))
		if (i < postdata size - 1, buffer appendSeq("&"))
	    )
	    content := buffer asString
	) else (
	    content := postdata
	)
	
	header appendSeq("Content-Length: ", content size, "\r\n\r\n", content)
	
	socket := Socket clone setHost(ip) setPort(port) 
	socket connect 
	
	if (socket error, 
	    error = socket error
	    return Nil
	)
	 
	socket write(header)
	while (socket read, socket readBuffer empty)
	socket close
	if (socket error, self setError(socket error); return Nil)     
	self
    )

    openOnDesktop := method(
	  platform := System platform
	  if( platform == "Mac OS/X" ) then (
	      System system( "open " .. url)
          ) elseif ( platform == "Windows NT" ) then (
	      System system( "cmd /c start " .. url)
          ) elseif ( platform == "Windows 9X" ) then (
	      System system( "command /c start " .. url)
          ) else (
             // assume generic Unix?
	      System system( "open " .. url)
          )
    )
)

Object doURL := method(url, self doString(URL clone setURL(url) fetch))
