/*
docCopyright("Steve Dekorte", 2002)
docLicense("BSD revised")
*/

#include "IoState.h"

#include "IoCLI.h"
#include "IoObject.h"
#include "IoObject_actor.h"

#include "IoException.h"
#include "IoFile.h"
#include "IoDirectory.h"
#include "IoList.h"
#include "IoNil.h"
#include "IoNumber.h"
#include "IoSeq.h"

int CLI_main(int argc, const char *argv[])
{
    IoCLI *cli = IoCLI_new();
    MainArgs_argc_argv_(cli->state->mainArgs, argc, argv);
    IoCLI_run(cli);
    IoCLI_free(cli);
    return 0;
}

IoCLI *IoCLI_new(void)
{
    IoCLI *self = calloc(1, sizeof(IoCLI));
    self->cli = CLI_new();
    return self;
}

void IoCLI_free(IoCLI *self)
{
    CLI_free(self->cli);
    free(self);
}

void IoCLI_state_(IoCLI *self, IoState *state) 
{ 
    self->state = state; 
    IoState_callbackContext_(self->state, self);
    IoState_printCallback_(self->state, (IoStatePrintCallback *)IoCLI_print_);  
}

int IoCLI_foundMainFile(IoCLI *self)
{
    int result;
    FILE *fp = fopen("main.io", "r");
    result = fp ? 1 : 0;
    
    if (fp) 
    {
	fclose(fp);
    }
    
    return result;
}

void IoCLI_printVersion(IoCLI *self)
{
    IoState *state = self->state;
    IoObject *lobby = state->lobby;
    IoNumber *version      = IoObject_getSlot_(lobby, IoState_symbolWithCString_(state, "version"));
    IoNumber *distribution = IoObject_getSlot_(lobby, IoState_symbolWithCString_(state, "distribution"));
    IoState_print_(state, "%s version %i\n", CSTRING(distribution), IoNumber_asInt(version));
}

void IoCLI_printHelp(IoCLI *self)
{
    printf("See: http://www.iolanguage.com/docs/\n");
}

void IoCLI_runCLI(IoCLI *self)
{
    IoCLI_printVersion(self);
    CLI_prompt_(self->cli, "Io> ");
    CLI_target_callback_(self->cli, self, (CLICallback *)IoCLI_cliInput);
    self->didPrint = 0;
    IoState_clearRetainStack(self->state);
    
    CLI_run(self->cli);
}

void IoCLI_runFile_argsList_(IoCLI *self, const char *fileName, IoList *argsList)
{ 
    IoState *state = self->state;
    IoObject *lobby = state->lobby;
    IoObject_setSlot_to_(lobby, IoState_symbolWithCString_(state, "args"), argsList);
    IoCLI_runFile_(self, (char *)fileName);
}

IoObject *IoCLI_run(IoCLI *self)
{
    IoState *state = self->state;
    IoObject *lobby = state->lobby;
    int argc = state->mainArgs->argc;
    const char **argv = state->mainArgs->argv;
    IoList *argsList = IoList_new(state);
    int i;
    int isInteractive = 0;
    
    /* this is a hack as it mangles mainArgs */
    if (argc >= 2 && (strcmp(argv[1], "-i") == 0))
    {
        int i;
        isInteractive = 1;

        for (i = 1; i < argc - 1; i ++)
        {
            argv[i] = argv[i + 1];
        }

        argc --;
    }
    
    for (i = 2; i < argc; i ++)
    { 
	IoList_rawAppend_(argsList, IoState_symbolWithCString_(state, (char *)argv[i])); 
    }
    
    /*
    int argc = 2;
    char **argv = malloc(2*sizeof(char *));
    argv[1] = "/Users/steve/Io/Io/IoVM/_tests/primitiveTests.io";
    */
    
    if (argc < 2) 
    {
	if (IoCLI_foundMainFile(self))
	{ 
	    IoCLI_runFile_argsList_(self, "main.io", argsList); 
	}
	else 
	{ 
	    ByteArray *path = IoDirectory_CurrentWorkingDirectoryAsByteArray();
	    IoObject_setSlot_to_(lobby, IoState_symbolWithCString_(state, "launchPath"), 
				 IoState_symbolWithByteArray_copy_(state, path, 0));    
	    
	    IoCLI_runCLI(self); 
	}
    }
    else if (strcmp(argv[1], "--help")==0)
    { 
	IoCLI_printHelp(self); 
    }
    else if (strcmp(argv[1], "--version")==0)
    { 
	IoCLI_printVersion(self); 
    }  
    else
    { 
	IoCLI_runFile_argsList_(self, argv[1], argsList); 
    }

    if (isInteractive) 
    {
	IoCLI_runCLI(self);
    }
    
    return state->ioNil;
}

void IoCLI_runFile_(IoCLI *self, const char *filePath)
{
    IoState *state = self->state;
    IoObject *lobby = state->lobby;
    ByteArray *path;
    
    /* --- set launch path --- */
    if (*filePath == '/' || *filePath == '\\')
    { 
	path = ByteArray_newWithCString_(filePath); 
    }
    else
    {
	path = IoDirectory_CurrentWorkingDirectoryAsByteArray();
	ByteArray_appendPathCString_(path, filePath);
    }
    
    ByteArray_removeLastPathComponent(path);  
    
    IoObject_setSlot_to_(lobby, 
			 IoState_symbolWithCString_(state, "launchPath"), 
			 IoState_symbolWithByteArray_copy_(state, path, 0));
    
    IoState_doFile_(state, filePath); 
    
    while (IoState_activeCoroCount(state) > 1)
    { 
	IoState_yield(state); 
    }
}

void IoCLI_cliInput(IoCLI *self, const char *s) 
{
    if (strcmp(s, "exit") == 0)
    {
	CLI_exit(self->cli);
	return;
    }
    else
    {
	IoObject *v = IoState_doCString_(self->state, s);
	IoObject_setSlot_to_(IOSTATE->lobby, IOSYMBOL("_"), v);
	
	if (self->didPrint) 
	{
	    IoState_justPrint_(self->state, "\n");
	}
	
	IoState_justPrint_(self->state, "==> ");
	
	if (v) 
	{
	    IoObject_print(v); 
	}
	
	IoState_justPrint_(self->state, "\n");
	IoState_clearRetainStack(self->state);
	self->didPrint = 0;
    }
}

/* --- callbacks -------------------------- */

void IoCLI_print_(IoCLI *self, const char *s)
{ 
    self->didPrint = 1;
    fputs(s, stdout);
}
