/*#io
Future ioDoc(
docCopyright("Steve Dekorte", 2002)
docLicense("BSD revised")    
docObject("Future")        
docDescription("A transparent container for a value returned by an asynchronous message. The future effectively becomes the return value when it is ready. If the future is sent a message before the asynchronous message is processed, the current coroutine will pause until the result is ready.")
*/

#include "IoFuture.h"
#include "IoObject.h"
#include "IoObject_actor.h"
#include "IoState.h"
#include "IoSeq.h"
#include "IoNil.h"

#define DATA(self) ((IoFutureData *)self->data)

IoTag *IoFuture_tag(void *state)
{
    IoTag *tag = IoTag_newWithName_("Future");
    tag->state = state;
    tag->cloneFunc = (TagCloneFunc *)IoFuture_rawClone;
    tag->markFunc  = (TagMarkFunc *)IoFuture_mark;
    tag->freeFunc  = (TagFreeFunc *)IoFuture_free;
    return tag;
}

IoFuture *IoFuture_proto(void *state)
{
    IoMethodTable methodTable[] = {
	{"forward", IoFuture_forward},
	{"isWaitingFuture", IoFuture_isWaitingFuture},
	/*
	  docSlot("?(<message>)")
	  docDescription("Same as Object ? operator.")
	*/
	{NULL, NULL},
    };

    IoObject *self = IoObject_new(state);

    self->data = calloc(1, sizeof(IoFutureData));
    self->tag = IoFuture_tag(state);
    /* rest of data expected to be NULL */
    IoState_registerProtoWithFunc_(state, self, IoFuture_proto);
      
    IoObject_addMethodTable_(self, methodTable);
    return self;
}

IoFuture *IoFuture_rawClone(IoFuture *proto) 
{ 
    IoFuture *self = IoObject_rawClonePrimitive(proto);
    self->data = cpalloc(proto->data, sizeof(IoFutureData));
    
    {
    /*
    PHash_copy_(self->slots, proto->slots);
    IoObject *isWaitingFuture = IoObject_getSlot_(proto, IOSYMBOL("isWaitingFuture"));
    IoObject *question = IoObject_getSlot_(proto, IOSYMBOL("?"));
    */

    IoObject *forward = IoObject_getSlot_(proto, IOSTATE->forwardString);


    IoObject_setSlot_to_(self, IOSTATE->forwardString, forward); /* so we have slots of our own */
    PHash_copy_(self->slots, proto->slots);
    /*
    IoObject_setSlot_to_(self, IOSYMBOL("isWaitingFuture"), isWaitingFuture);
    IoObject_setSlot_to_(self, IOSYMBOL("?"), question);
    */

    IoObject_rawRemoveAllProtos(self);
    }    
    return self; 
}

IoFuture *IoFuture_new(void *state)
{
    IoFuture *proto = IoState_protoWithInitFunction_(state, IoFuture_proto);
    return IOCLONE(proto);
}

void IoFuture_free(IoFuture *self)
{
    /*IoFuture *proto = IoState_protoWithInitFunction_(IOSTATE, IoFuture_proto);*/
    if (NULL == DATA(self)) 
    { 
	return; 
    }
    
    free(DATA(self));
    self->data = NULL;
}


IoFuture *IoFuture_newWithMessage_creatorObject_(IoState *state, 
    IoMessage *message, 
    IoObject *creatorObject)
{
    IoFuture *self = IoFuture_new(state);
    DATA(self)->message = IOREF(message);
    DATA(self)->creatorObject = IOREF(creatorObject);
    return self;
}

IoMessage *IoFuture_rawMessage(IoFuture *self)
{ 
    return DATA(self)->message; 
}

void IoFuture_mark(IoFuture *self)
{ 
    if (DATA(self)->creatorObject) 
    {
	IoObject_makeGrayIfWhite((IoObject *)(DATA(self)->creatorObject));
    }
    
    if (DATA(self)->pausedObject) 
    {
	IoObject_makeGrayIfWhite((IoObject *)(DATA(self)->pausedObject));
    }
    
    if (DATA(self)->message) 
    {
	IoObject_makeGrayIfWhite((IoObject *)(DATA(self)->message));
    }
}

void IoFuture_rawSetValue_(IoFuture *self, IoObject *v) 
{ 
    IoObject *pausedObject = DATA(self)->pausedObject;

    /* a little magic to turn the future into the result value */
    /* we empty the slots and set the proto slot to the result */
    IoObject_clean(self);
    
    /* IoObject_removeSlot_(self, IOSTATE->forwardString); */
    IoObject_rawSetProto_(self, v);
    
    /* 
    set the tag and data to the result tag and data 
    need to add something to prevent proto slot from changing 
    in order to avoid freeing the data while we still need it 
    */
    self->tag = v->tag;
    free(DATA(self));
    self->data = v->data;
    self->extraData = v->extraData;
    self->mark.doesNotOwnData = 1;

    /* then we can resume the paused object */
    if (pausedObject) 
    {
	IoObject_resume(pausedObject);
    }
}

unsigned char IoFuture_isPausing_(IoFuture *self, IoObject *v)
{
    if (DATA(self)->pausedObject) 
    { 
	return IoObject_isPausing_(DATA(self)->pausedObject, v); 
    } 
    
    return 0;
}

/* ----------------------------------------------------------- */

IoObject *IoFuture_setValue(IoFuture *self, IoObject *locals, IoMessage *m)
{ 
    /*
    undocumented: yes
    method: 
    description:
    */
    
    IoObject *v = IoMessage_locals_valueArgAt_(m, locals, 0);
    IoFuture_rawSetValue_(self, v); 
    return self; 
}

inline int IoFuture_rawIsReady(IoFuture *self)
{ 
    return IoObject_hasProtos(self); 
}

IoObject *IoFuture_isWaitingFuture(IoFuture *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("isWaitingFuture", 
    "Returns Nop. ")
    */
    
    return IOSTATE->ioNop; 
}

int IoFuture_rawWaitOnResult(IoFuture *self)
{     
    if (IoFuture_rawIsReady(self)) 
    {
	return 0;
    }
    
    IoState_yield(IOSTATE);

    if (IoFuture_rawIsReady(self)) 
    {
	return 0;
    }
    
    /* auto deadlock detection */
    if (IoObject_isPausing_(IoState_currentActor(IOSTATE), DATA(self)->creatorObject))
    { 
	return -1;
    }
    
    DATA(self)->pausedObject = IOREF(IoState_currentActor(IOSTATE));
    IoObject_pause(DATA(self)->pausedObject); /* this doesn't work on the root coro? */

    while (!IoFuture_rawIsReady(self)) 
    { 
	IoObject_yield(DATA(self)->pausedObject); 
    } 
    
    /* in case it's the root coro? */
    return 0;
}

IoObject *IoFuture_forward(IoFuture *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("forward", 
    "Implements Future's waiting behavior. ")
    */
    
    if (IoFuture_rawWaitOnResult(self) == -1)
    {
	IoState_error_description_(IOSTATE, m, 
	    "Io Future value", 
	    "attempt to block while waiting for this Future result would create a deadlock");
	return IONIL(self);
    }
    
    return IoObject_perform(self, locals, m);
}

