/*#io
Scheduler ioDoc(
docCopyright("Steve Dekorte", 2002)
docLicense("BSD revised")
docObject("Scheduler")
docDescription("Contains methods related to the IoVM coroutine scheduler.")
*/

#if defined(unix) || defined(__APPLE__)
#include <sys/utsname.h>
#endif

#include "IoScheduler.h"
#include "IoNil.h"
#include "IoNumber.h"
#include "IoList.h"

#if defined(__CYGWIN__) || defined(_WIN32)
#include <windows.h>
#endif

IoObject *IoScheduler_proto(void *state)
{
    IoMethodTable methodTable[] = 
    { 
	{"actors", IoObject_actors},
	{"currentActor", IoObject_currentActor},
	{"activeActorCount", IoObject_activeActorCount},
	{"setSleepInterval", IoObject_setSleepInterval},
	{"sleepInterval", IoObject_sleepInterval},
	{0x0, 0x0},
    };
    IoObject *self = IoObject_new(state);
    IoObject_addMethodTable_(self, methodTable);
    return self;
}

IoObject *IoObject_actors(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("actors", 
    "Returns a copy of the list of all active actors.")
    */
    
    IoState *state = IOSTATE;
    IoList *ioList = IoList_new(state);
    List *list = IoList_rawList(ioList);
    List_copy_(list, state->actors);
    return ioList; 
}

IoObject *IoObject_currentActor(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("currentActor", 
    "Returns the current actor object.")
    */
    
    return IoState_currentActor(IOSTATE); 
}

IoObject *IoObject_activeActorCount(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("activeActorCount", 
    "Returns the number of active actors.")
    */
    
    return IONUMBER(Scheduler_coroCount(IOSTATE->scheduler));
}

IoObject *IoObject_setSleepInterval(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("setSleepInterval(aNumber)", 
    "Sets the sleep interval (in seconds) per scheduler cycle. 
    The actual amount of time slept each yield is this interval divided by the
    number of active actors at the moment of the yield. In this way, the total 
    time slept per scheduler cycle should typically be close to the sleepInterval. 
    Return self.")
    */
    
    double seconds = IoMessage_locals_doubleArgAt_(m, locals, 0);
    
    if (seconds >= 1)
    {
	IoState_error_description_(IOSTATE, m,
	    "Io System Scheduler setSleepSeconds", 
	    "period of '%f' is too big - must be less than 1", seconds);  
    }
    
    Scheduler_secondsInSleepPeriod_(IOSTATE->scheduler, (float)seconds);
    return self;
}

IoObject *IoObject_sleepInterval(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("sleepInterval", 
    "Returns a number containing the sleep interval (in seconds)
    between actor coroutine switches. ")
    */
    
    return IONUMBER((double)Scheduler_secondsInSleepPeriod(IOSTATE->scheduler));
}
