/*#io
docCopyright("Steve Dekorte", 2002)
docLicense("BSD revised")
*/

#ifndef IOSTATE_DEFINED 
#define IOSTATE_DEFINED 1

#define SIOSYMBOL(stringConstant) IoState_symbolWithCString_(self, stringConstant)

#define IOASSERT(value, message) \
  if (!(value)) \
  { IoState_error_description_(IOSTATE, m, "Io Assertion", message); }

/*#define IOSTATE ((IoState *)(self->tag->state))*/
#define IOSTATE ((IoState *)(self->state))
#define IONOP(self) (IOSTATE->ioNop)

#include "Queue.h"
#include "Stack.h"
#include "Hash.h"
#include "MainArgs.h"
#include "Scheduler.h"
#include "IoObject_struct.h"
#include "IoSeq.h"
#include "IoObjectGroup.h"
#include "IoVersion.h"
#include "IoException.h"
#include "IoStore.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef void (IoStateBindingsInitCallback)(void *, void *);
typedef void (IoStatePrintCallback)(void *, const char *);
typedef void (IoStateExceptionCallback)(void *, IoException *);
typedef void (IoStateExitCallback)(void *, int);
typedef void (IoStateActiveCoroCallback)(void *, int);
typedef IoObject *(IoStateProtoFunc)(void *);

typedef struct IoState IoState;

struct IoState
{
    Hash *primitiveProtos;

    SkipDBM *sdbm;
    SkipDB *symbols;

    /* -- actors --- */
    Scheduler *scheduler; 
    List *actors;		/* to prevent active actors from being collected */
    IoObject *mainActor;	/* the object that represents the main "thread" */
    IoObject *currentActor;	/* the object whose coroutine is active */
    Stack *currentRetainStack;	/* quick access to current coro's retain stack */

    /* --- quick access objects --- */
    IoSymbol *activateString;
    IoSymbol *thisBlockString;
    IoSymbol *cloneString;
    IoSymbol *CompilerString;
    IoSymbol *debugString;
    IoSymbol *debugParserString;
    IoSymbol *delegateString;
    IoSymbol *equalEqualString;
    IoSymbol *forwardString;
    IoSymbol *initString;
    IoSymbol *thisMessageString;
    IoSymbol *objectString;
    IoSymbol *parentString;
    IoSymbol *returnString;
    IoSymbol *selfString;
    IoSymbol *senderString;
    IoObject *setSlotBlock;
    IoSymbol *setSlotString;
    IoSymbol *slotContextString;
    IoSymbol *targetString;
    IoSymbol *updateSlotString;
    IoObject *localsUpdateSlotCFunc;
    IoObject *localsProto;

    IoMessage *initMessage;
    IoMessage *compareMessage;
    IoMessage *nilMessage;
    IoMessage *printMessage;

    /* --- global objects --- */
    IoObject *ioNil; /* object that is no other object */
    IoObject *ioNop; /* object that returns itself for any message */

    IoObject *errorException;
    IoObject *warningException;
    IoObject *notificationException;

    /* -- garbage collector --- */
    IoObject *lobby;
    List *retainedValues;

    int collectorPauseCount;
    /*int sweepSize;*/ /* number of objects until next sweep */

    IoObjectGroup *blackGroup;
    IoObjectGroup *grayGroup;
    IoObjectGroup *whiteGroup;
    IoObjectGroup *freeGroup;
    
    unsigned int currentBlackColor;
    unsigned int currentWhiteColor;

    size_t sweepsPerGeneration;
    size_t sweepsSinceLastGeneration;
    
    size_t newValueCount;             /* count of values added since last mark phase */
    size_t newValueCountToUpdate;     /* do a mark when reached */
    size_t newValueSinceSweepCount;   /* count of values added since last sweep phase */
    size_t allocsPerSweep;		   /* do a sweep when reached */
    size_t lastReferencedObjectCount; /* number of referenced objects (after sweep) */

    /* -- weaklinks --- */
    Hash *valueToWeakLink;
    /*Stack *collectedWeakLinks;*/

    /* -- recycling --- */
    List *recycledObjects;
    List *recycledObjectsSansSlots;

    MainArgs *mainArgs;

    /* --- current execution state --- */
    int stopStatus;
    void *returnValue;

    /* -- embedding --- */
    void *callbackContext;
    IoStateBindingsInitCallback *bindingsInitCallback;
    IoStatePrintCallback        *printCallback;
    IoStateExceptionCallback    *exceptionCallback;
    IoStateExitCallback         *exitCallback;
    IoStateActiveCoroCallback   *activeCoroCallback;

    /* --- persistence --- */
    IoStore *store; 
    int isCleaningUp; /* set to 1 when IoState_free is called */
    int debugCollector;
    IoObject *debugger;
    IoMessage *vmWillSendMessage;

    /* --- SandBox Limits --- */
    int messageCountLimit;
    int messageCount;
    
    double timeLimit;
    double endTime;
    IoMessage *tailCallMessage;
};

#define IOSTATE_STRUCT_DEFINED

/* --- setup ---*/
IoState *IoState_new(void);
void IoState_init(IoState *self);

/* --- setup tags ---*/
void IoState_registerProtoWithFunc_(IoState *self, IoObject *proto, IoStateProtoFunc *func);
IoObject *IoState_protoWithInitFunction_(IoState *self, IoStateProtoFunc *func);
IoObject *IoState_protoWithName_(IoState *self, const char *name);

void IoState_free(IoState *self);
IoObject *IoState_lobby(IoState *self);
void IoState_setLobby_(IoState *self, IoObject *obj);

/* --- performing --------------------------------- */

IoObject *IoState_doCString_(IoState *self, const char *s);
IoObject *IoState_doSandboxCString_(IoState *self, const char *s);
IoObject *IoState_doFile_(IoState *self, const char *fileName);

/* --- embedding callbacks ------------------------ */


void IoState_callbackContext_(IoState *self, void *context);
void *IoState_callbackContext(IoState *self);

void IoState_setBindingsInitCallback(IoState *self, IoStateBindingsInitCallback *callback);

void IoState_print_(IoState *self, const char *format, ...);
void IoState_justPrint_(IoState *self, const char *s);
void IoState_printCallback_(IoState *self, IoStatePrintCallback *callback);

void IoState_exceptionCallback_(IoState *self, IoStateExceptionCallback* callback);
void IoState_exception_(IoState *self, IoException *e);

void IoState_exitCallback_(IoState *self, IoStateExitCallback* callback);
void IoState_exit(IoState *self, int returnCode);

/* --- active coro callback --------------------------- */

/* called when coro count changes */
void IoState_activeCoroCallback_(IoState *self, IoStateActiveCoroCallback* callback);
void IoState_schedulerUpdate(IoState *self, int count);

/* --- actors ------------------------------------- */
void IoState_addActor_(IoState *self, IoObject *actor);
void IoState_removeActor_(IoState *self, IoObject *actor);
IoObject *IoState_currentActor(IoState *actor);
void IoState_currentActor_(IoState *self, IoObject *actor);
int IoState_activeCoroCount(IoState *self);

void IoState_yield(IoState *self);
void *IoState_prepareToPause(IoState *self);
void IoState_resume_(IoState *self, IoObject *actor);

/* --- command line ------------------------------- */
void IoState_argc_argv_(IoState *st, int argc, const char *argv[]);
/*int IoState_main(IoState *self, int argc, const char **argv);*/

/* --- Store --------------------------------------------------- */
IoStore *IoState_store(IoState *self);

/* --- debugging --------------------------------------------------- */
void IoState_show(IoState *self);

IoObject *IoState_replacePerformFunc_with_(IoState *self, 
    TagPerformFunc *oldFunc, 
    TagPerformFunc *newFunc);
    
void IoState_debuggingOn(IoState *self);
void IoState_debuggingOff(IoState *self);
int IoState_hasDebuggingActor(IoState *self);
void IoState_updateDebuggingMode(IoState *self);

double IoState_endTime(IoState *self);

#include "IoState_collector.h"
#include "IoState_inline.h"
#include "IoState_exceptions.h"
#include "IoState_symbols.h"

#ifdef __cplusplus
}
#endif
#endif
